'''

Created on Dec 22, 2020
@author: ar-calder

'''

from datetime import datetime, timedelta
import dateutil.parser
import json
import logging
import re

logger = logging.getLogger(__name__)


def iso8601_to_date(iso_string, with_zone=False):
    """Utility function to convert iso_8601 formatted string to datetime object, optionally accounting for timezone

    Args:
        iso_string (string): the iso_8601 string to convert to datetime object
        with_zone (bool, optional): whether to account for timezone offset. Defaults to False.

    Returns:
        datetime.datetime: equivalent time, with or without timezone offsets
    """
    date_timezone = iso_string.split('Z')
    date = dateutil.parser.parse(date_timezone[0])
    if with_zone and len(date_timezone > 1):
        hours_minutes = date_timezone[1].split(':')
        minutes = (60*int(hours_minutes[0]) + int(hours_minutes[1] if len(hours_minutes) > 1 else 0))
        date = date + datetime.timedelta(minutes=minutes)
    return date

def iso8601_timespan(days_ago, from_date=datetime.utcnow(), delta=timedelta(weeks=1)):
    curr_date = from_date - timedelta(days=days_ago)
    while curr_date < from_date:
        yield curr_date.isoformat('T', 'seconds')
        curr_date += delta

def min_iso8601():
    """Utility wrapper for iso8601_to_date which provides minimum date (for comparison purposes).

    Returns:
        datetime.datetime: 0 / 1970-01-01T00:00:00.000
    """
    return iso8601_to_date("1970-01-01T00:00:00.000")

def find_field(data_to_filter, field_name, field_value):
    """Utility function to filter blackduck objects for specific fields

    Args:
        data_to_filter (dict): typically the blackduck object or subselection of this
        field_name (string): name of field to use in comparisons
        field_value (string): value of field we seek

    Returns:
        object: object if found or None.
    """
    return next(filter(lambda d: d.get(field_name) == field_value, data_to_filter), None)

def safe_get(obj, *keys):
    """Utility function to safely perform multiple get's on a dict.
       Particularly useful on complex/deep objects.

    Args:
        obj (dict): object to perform get on.
        *keys (string): consecutive keys as args.

    Returns:
        object: object if found or None.
    """
    for key in keys:
        try:
            obj = obj[key]
        except KeyError:
            return None
    return obj

def get_url(obj):
    """Utility wrapper for safe_get providing URL lookup for a given object

    Args:
        obj (dict): object to perform URL lookup on.

    Returns:
        string: url if found or None.
    """
    return safe_get(obj, '_meta', 'href')

def get_resource_name(obj):
    """Utility function to determine resource name from a given resource object

    Args:
        obj (dict): object to perform name lookup on.

    Returns:
        string: name if found or None.
    """
    parts = get_url(obj).split('/')
    print("parts =", get_url(obj))
    for part in reversed(parts[:-1]):
        # regex for id 8-4-4-12
        if re.search("^\w{8}-\w{4}-\w{4}-\w{4}-\w{12}$", part):
            continue    
        return part
    

def pfmt(value):
    """Utility function to 'pretty format' a dict or json 

    Args:
        value (json/dict): the json object or dict to pretty format

    Returns:
        string: json formatted string representing passed object
    """
    return json.dumps(value, indent=4)

def pprint(value):
    """Utility wrapper for pfmt that prints 'pretty formatted' json data.

    Args:
        value (json/dict): the json object or dict to pretty print

    Returns:
        None
    """
    print(pfmt(value))

def object_id(object):
    assert '_meta' in object, "REST API object must have _meta key"
    assert 'href' in object['_meta'], "REST API object must have href key in it's _meta"
    return object['_meta']['href'].split("/")[-1]

def expect_type(given, expected):
    """Utility wrapper for assert isinstance.

    Args:
        given (object): object to compare
        expected (type): expected object type

    Throws:
        AssertionError: on expected type != given type
    """
    assert isinstance(given, expected), f"Expected {expected} given {type(given)}"




"""
Created on Dec 23, 2020
@author: ar-calder

Wrapper for common HUB API queries. 
Upon initialization Bearer token is obtained and used for all subsequent calls.
Token will auto-renew on timeout.
"""

from .Utils import safe_get
from .Authentication import BearerAuth
import json
import logging
import os
from pprint import pformat
import requests.packages.urllib3
from requests.packages.urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter
from urllib.parse import urljoin

logger = logging.getLogger(__name__)


class HubSession(requests.Session):
    """Hold base_url, timeout, retries, and provide sensible defaults"""

    def __init__(self, base_url, timeout, retries, verify):
        super().__init__()
        self.base_url = base_url
        self._timeout = float(timeout)  # timeout is not a member of requests.Session
        self.verify = verify

        # use sane defaults to handle unreliable networks
        """HTTP response status codes:
                429 = Too Many Requests
                500 = Internal Server Error
                502 = Bad Gateway
                503 = Service Unavailable
                504 = Gateway Timeout
        """
        retry_strategy = Retry(
            total=int(retries),
            backoff_factor=2,  # exponential retry 1, 2, 4, 8, 16 sec ...
            status_forcelist=[429, 500, 502, 503, 504],
            method_whitelist=['GET']
        )

        adapter = HTTPAdapter(max_retries=retry_strategy)
        self.mount("https://", adapter)
        self.mount("http://", adapter)
        logger.info("Using a session with a %s second timeout and up to %s retries per request", timeout, retries)

        self.proxies.update({
            'http': os.environ.get('http_proxy', ''),
            'https': os.environ.get('https_proxy', '')
        })

    def request(self, method, url, **kwargs):
        kwargs['timeout'] = self._timeout

        if method.lower() == 'get':
            headers = kwargs.pop('headers', dict())
            lc_keys = {key.lower(): value for (key, value) in headers.items()}
            if 'accept' not in lc_keys and 'content-type' not in lc_keys:
                # set default media type only if neither 'accept' nor 'content-type'
                # exist as some endpoints may only accept one or the other but not both
                lc_keys['accept'] = "application/json"
                lc_keys['content-type'] = "application/json"
            kwargs['headers'] = lc_keys

        url = urljoin(self.base_url, url)
        return super().request(method, url, **kwargs)


class Client:
    """A binding to Blackduck's REST API that provides a robust connection backed by a session object.
    A base URL, timeout, retries, proxies, and TLS verification are set upon initialization and these
    attributes are persisted across all requests.

    At the REST API level, it provides a consistent way to discover and traverse public resources,
    uses a generator to fetch all items using pagination, and automatically renews the bearer token.

    Ultimately it provides a solid foundation especially suited for long-running scripts.
    """
    def __init__(self,
                 token=None,
                 base_url=None,
                 session=None,
                 auth=None,
                 verify=True,
                 timeout=15.0,  # in seconds
                 retries=3):
        """Instantiate a Client for use with Hub's REST-API

        Args:
            token (str): Access Token obtained from the Hub UI: System -> My Access Tokens
            base_url (str): e.g. "https://your.blackduck.url"
            session (requests.Session): custom session if specified.  For advanced users only.
                If not provided, a HubSession with recommended defaults will be generated and used.
                Any custom session must incorporate a base_url in every request as a plain
                requests.Session() will not work.  See HubSession implementation for an example.
            auth (requests.auth.AuthBase): custom authorization if specified. For advanced users only.
                If not provided, one based on the access token is generated and used.
            verify (bool): TLS certificate verification. Defaults to True.
            timeout (float): request timeout in seconds. Defaults to 15 seconds.
            retries (int): maximum number of times to retry a request. Defaults to 3.
        """
        self.base_url = base_url
        self.session = session or HubSession(base_url, timeout, retries, verify)
        self.session.auth = auth or BearerAuth(self.session, token)
        self.root_resources_dict = None

    def list_resources(self, parent=None):
        """List named resources that can be fetched.

        Args:
            parent (dict/json): resource object from prior get_resource invocations.
                                Defaults to None (for root /api/ base).

        Returns:
            dict(str -> str): of public resource names to urls
                              To obtain the url to the parent itself, use key 'href'.
        """
        if parent is not None and not isinstance(parent, dict):
            raise TypeError("parent parameter must be a dict if not None")

        if not parent:
            # the root resources are in a different format (name -> href)
            # compared to (rel, href) pairs in _meta.links
            if self.root_resources_dict is None:
                # cache root resources for efficiency
                resp = self.session.get("/api/")
                resources_dict = resp.json()
                resources_dict['href'] = resp.url  # save url to root itself
                del resources_dict['_meta']
                self.root_resources_dict = resources_dict
            return self.root_resources_dict
        else:
            key = '_hub_rest_api_python_resources_dict'
            if key not in parent:
                obj = safe_get(parent, '_meta', 'links')
                try:
                    rel_href_pairs = iter(obj)
                except TypeError:
                    logger.error("unable to list resources on parent object (missing ['_meta']['links']):")
                    logger.error(pformat(parent))
                    raise
                resources_dict = {}
                for res in rel_href_pairs:
                    resources_dict[res['rel']] = res['href']
                # save url to parent itself if available, otherwise save 'href': None
                resources_dict['href'] = safe_get(parent, '_meta', 'href')
                parent[key] = resources_dict  # cache for future use
            return parent[key]

    def get_resource(self, name, parent=None, items=True, **kwargs):
        """Fetch a named resource.

        Args:
            name (str): resource name i.e. specific key from list_resources()
            parent (dict/json): resource object from prior get_resource() call.
                                Use None for root /api/ base.
            items (bool): enable resource generator for paginated results. Defaults to True.
            kwargs: passed to session.request

        Returns:
            list (items=True) or dict formed from returned json
        """
        if not isinstance(name, str) or not name:
            raise TypeError("name parameter must be a non-empty str")
        if parent is not None and not isinstance(parent, dict):
            raise TypeError("parent parameter must be a dict if not None")

        resources_dict = self.list_resources(parent)
        if name not in resources_dict:
            msg = f"resource name '{name}' not found in available resources"
            logger.error(msg)
            logger.error(pformat(resources_dict))
            raise KeyError(msg)
        url = resources_dict[name]

        if items:
            return self.get_items(url, **kwargs)
        else:
            return self.get_json(url, **kwargs)

    def get_metadata(self, name, parent=None, **kwargs):
        """Fetch named resource metadata and other useful data such as totalCount.

        Args:
            name (str): resource name i.e. specific key from list_resources()
            parent (dict/json): resource object from prior get_resource() call.
                                Use None for root /api/ base.
            kwargs: passed to session.request

        Returns:
            dict/json: named resource metadata
        """
        # limit: 0 works for 'projects' but not for 'codeLocations' or project 'versions'
        kwargs['params'] = {'limit': 1}
        return self.get_resource(name, parent, items=False, **kwargs)

    def get_json(self, url, **kwargs):
        """Streamline GET request to url endpoint and return json result
           while preserving underlying error handling.

        Args:
            url (str): of endpoint
            kwargs: passed to session.request

        Returns:
            json/dict: requested object

        Raises:
            requests.exceptions.HTTPError: from response.raise_for_status()
            json.JSONDecodeError: if response.text is not json
        """
        r = self.session.get(url, **kwargs)

        if r.status_code != 200:
            # print out a more descriptive error message before raising an exception
            self.http_error_handler(r)

        r.raise_for_status()

        if 'Content-Type' in r.headers:
            content_type = r.headers['Content-Type']
            if 'internal' in content_type:
                logger.warning("Response contains internal proprietary Content-Type: " + content_type)

        try:
            return r.json()
        except json.JSONDecodeError:
            self.http_error_handler(r)
            raise

    def get_items(self, url, page_size=100, **kwargs):
        """Fetch 'pages' of items

        Args:
            url (str): of endpoint
            page_size (int): Number of items to get per page. Defaults to 100.
            kwargs: passed to session.request

        Yields:
            generator(dict/json): of items
        """
        offset = 0
        params = kwargs.pop('params', dict())

        while True:
            params.update({'offset': f"{offset}", 'limit': f"{page_size}"})
            kwargs['params'] = params
            items = self.get_json(url, **kwargs).get('items', list())

            for item in items:
                yield item

            if len(items) < page_size:
                # This will be true if there are no more 'pages' to view
                break

            offset += page_size

    @staticmethod
    def http_error_handler(r):
        """Handle an unexpected HTTPError or Response by logging useful information.

        Args:
            r (requests.HTTPError OR requests.Response): to handle
        """
        if isinstance(r, requests.HTTPError):
            r = r.response
        logger.error(f"{r.request.method} {r.url}")
        status_description = requests.status_codes._codes[r.status_code][0]
        logger.error(f"HTTP response status code {r.status_code}: {status_description}")
        try:
            content = json.dumps(r.json(), indent=4)
            logger.error(f"HTTP response json (formatted): {content}")
        except json.JSONDecodeError:
            logger.error(f"HTTP response text: {r.text}")



import logging
import requests
import json
from operator import itemgetter
import urllib.parse

logger = logging.getLogger(__name__)

def upload_scan(self, filename):
    url = self.get_apibase() + "/scan/data/?mode=replace"
    headers = self.get_headers()
    if filename.endswith('.json') or filename.endswith('.jsonld'):
        headers['Content-Type'] = 'application/ld+json'
        with open(filename,"r") as f:
            response = requests.post(url, headers=headers, data=f, verify=not self.config['insecure'])
    elif filename.endswith('.bdio'):
        headers['Content-Type'] = 'application/vnd.blackducksoftware.bdio+zip'
        with open(filename,"rb") as f:
            response = requests.post(url, headers=headers, data=f, verify=not self.config['insecure'])
    else:
        raise Exception("Unkown file type")
    return response

def download_project_scans(self, project_name,version_name, output_folder=None):
    version = self.get_project_version_by_name(project_name,version_name)
    codelocations = self.get_version_codelocations(version)
    import os
    if output_folder:
        if not os.path.exists(output_folder):
            os.makedirs(output_folder, 0o755, True)
    
    result = []
    
    for item in codelocations['items']:
        links = item['_meta']['links']
        matches = [x for x in links if x['rel'] == 'enclosure' or x['rel'] == 'scan-data']
        for m in matches:
            url = m['href']
            filename = url.split('/')[6]
            if output_folder:
                pathname = os.path.join(output_folder, filename)
            else:
                if not os.path.exists(project_name):
                    os.mkdir(project_name)
                pathname = os.path.join(project_name, filename)
            responce = requests.get(url, headers=self.get_headers(), stream=True, verify=not self.config['insecure'])
            with open(pathname, "wb") as f:
                for data in responce.iter_content():
                    f.write(data)
            result.append({filename, pathname})
    return result
                        
def get_codelocations(self, limit=100, unmapped=False, parameters={}):
    parameters['limit'] = limit
    paramstring = self._get_parameter_string(parameters)
    headers = self.get_headers()
    url = self.get_apibase() + "/codelocations" + paramstring
    headers['Accept'] = 'application/vnd.blackducksoftware.scan-4+json'
    response = requests.get(url, headers=headers, verify = not self.config['insecure'])
    jsondata = response.json()
    if unmapped:
        jsondata['items'] = [s for s in jsondata['items'] if 'mappedProjectVersion' not in s]
        jsondata['totalCount'] = len(jsondata['items'])
    return jsondata

def get_codelocation_scan_summaries(self, code_location_id = None, code_location_obj = None, limit=100):
    '''Retrieve the scans (aka scan summaries) for the given location. You can give either
    code_location_id or code_location_obj. If both are supplied, precedence is to use code_location_obj
    '''
    assert code_location_id or code_location_obj, "You must supply at least one - code_location_id or code_location_obj"

    paramstring = "?limit={}&offset=0".format(limit)
    headers = self.get_headers()
    headers['Accept'] = 'application/vnd.blackducksoftware.scan-4+json'
    if code_location_obj:
        url = self.get_link(code_location_obj, "scans")
    else:
        url = self.get_apibase() + \
            "/codelocations/{}/scan-summaries".format(code_location_id)
    response = requests.get(url, headers=headers, verify = not self.config['insecure'])
    jsondata = response.json()
    return jsondata

def delete_unmapped_codelocations(self, limit=1000):
    code_locations = self.get_codelocations(limit=limit, unmapped=True).get('items', [])

    for c in code_locations:
        scan_summaries = self.get_codelocation_scan_summaries(code_location_obj = c).get('items', [])

        if scan_summaries[0]['status'] == 'COMPLETE':
            response = self.execute_delete(c['_meta']['href'])

def delete_codelocation(self, locationid):
    url = self.config['baseurl'] + "/api/codelocations/" + locationid
    headers = self.get_headers()
    response = requests.delete(url, headers=headers, verify = not self.config['insecure'])
    return response
    
def get_scan_locations(self, code_location_id):
    headers = self.get_headers()
    headers['Accept'] = 'application/vnd.blackducksoftware.scan-4+json'
    url = self.get_apibase() + "/codelocations/{}".format(code_location_id)
    response = requests.get(url, headers=headers, verify = not self.config['insecure'])
    jsondata = response.json()
    return jsondata



import logging
import requests
import json
from operator import itemgetter
import urllib.parse

from .Exceptions import InvalidVersionPhase

logger = logging.getLogger(__name__)

# VERSION_DISTRIBUTION=["EXTERNAL", "SAAS", "INTERNAL", "OPENSOURCE"]
# VERSION_PHASES = ["PLANNING", "DEVELOPMENT", "PRERELEASE", "RELEASED", "DEPRECATED", "ARCHIVED"]
# PROJECT_VERSION_SETTINGS = ['nickname', 'releaseComments', 'versionName', 'phase', 'distribution', 'releasedOn']
from .constants import VERSION_DISTRIBUTION, VERSION_PHASES, PROJECT_VERSION_SETTINGS

def _get_projects_url(self):
    return self.get_urlbase() + "/api/projects"

def get_projects(self, limit=100, parameters={}):
    headers = self.get_headers()
    if limit:
        parameters.update({'limit': limit})
    url = self._get_projects_url() + self._get_parameter_string(parameters)
    headers['Accept'] = 'application/vnd.blackducksoftware.project-detail-4+json'
    logger.debug(f"Retrieving projects using url {url}")
    response = requests.get(url, headers=headers, verify = not self.config['insecure'])
    jsondata = response.json()
    return jsondata

def create_project(self, project_name, version_name="Default Version", parameters={}):
    url = self._get_projects_url()

    post_data = {
        "name": project_name,
        "description": parameters.get("description", ""),
        "projectTier": parameters.get("project_tier", ""),
        "projectOwner": parameters.get("project_owner", ""),
        "projectLevelAdjustments": parameters.get("project_level_adjustments", True),
        "cloneCategories": [
        "COMPONENT_DATA",
        "VULN_DATA"
        ],
        "versionRequest": {
        "phase": parameters.get("version_phase", "PLANNING"),
        "distribution": parameters.get("version_distribution", "EXTERNAL"),
        "projectLevelAdjustments": parameters.get("project_level_adjustments", True),
        "versionName": version_name
        }
    }
    response = self.execute_post(url, data=post_data)
    return response

def create_project_version(self, project_obj, new_version_name, clone_version=None, parameters={}):
    url = self.get_link(project_obj, "versions")

    version_phase = parameters.get("phase", "PLANNING")
    if version_phase not in VERSION_PHASES:
        raise InvalidVersionPhase("The phase given {} is not in the list of valid phases ({})".format(
            version_phase, VERSION_PHASES))

    post_data = {
        "versionUrl": url,
        "cloneCategories": [
            "VULN_DATA",
            "COMPONENT_DATA"
        ],
        "versionName": new_version_name,
        "phase": version_phase,
        "distribution": parameters.get("distribution", "EXTERNAL")
    }
    if clone_version:
        post_data["cloneFromReleaseUrl"] = clone_version['_meta']['href']
    response = self.execute_post(url, data=post_data)
    return response

def get_project_by_name(self, project_name):
    project_list = self.get_projects(parameters={"q":"name:{}".format(project_name)})
    for project in project_list['items']:
        if project['name'] == project_name:
            return project

def get_projects_by_version_name(self, version_name, exclude_projects=None):
    """Returns all project dicts which have given version_name, including the version object under 'version' key
    
    Arguments:
        version_name {str} -- version name to be searched
        exclude_projects {list} -- list of project names to be excluded from scanning for given version name
    """
    headers = self.get_headers()
    projects = self.get_projects(limit=9999).get('items',[])
    if len(projects) == 0:
        logger.error('No projects found')
    else:
        jsondata = {'items':[]}
        for project in projects:
            if project['name'] not in exclude_projects:
                version = self.get_version_by_name(project, version_name)
                if version:
                    project['version'] = version
                    jsondata['items'].append(project)
        jsondata['totalCount'] = len(jsondata['items'])
        return jsondata

def get_version_by_name(self, project, version_name):
    version_list = self.get_project_versions(project, parameters={'q':"versionName:{}".format(version_name)})
    # A query by name can return more than one version if other versions
    # have names that include the search term as part of their name
    for version in version_list['items']:
        if version['versionName'] == version_name:
            return version

def get_project_version_by_name(self, project_name, version_name):
    project = self.get_project_by_name(project_name)
    if project:
        version = self.get_version_by_name(project, version_name)
        if version == None:
            logger.debug("Did not find any project version matching {}".format(version_name))
        else:
            return version
    else:
        logger.debug("Did not find a project with name {}".format(project_name))

def get_or_create_project_version(self, project_name, version_name, parameters = {}):
    project = self.get_project_by_name(project_name)
    if project:
        version = self.get_version_by_name(project, version_name)
        if not version:
            self.create_project_version(project, version_name, parameters=parameters)
            version = self.get_version_by_name(project, version_name)
    else:
        self.create_project(project_name, version_name, parameters=parameters)
        project = self.get_project_by_name(project_name)
        version = self.get_version_by_name(project, version_name)
    return version

def get_project_by_id(self, project_id, limit=100):
    headers = self.get_headers()
    paramstring = self.get_limit_paramstring(limit)
    url = self._get_projects_url() + "/" + project_id + paramstring
    headers['Accept'] = 'application/vnd.blackducksoftware.project-detail-4+json'
    response = requests.get(url, headers=headers, verify = not self.config['insecure'])
    jsondata = response.json()
    return jsondata

def get_project_versions(self, project, limit=100, parameters={}):
    # paramstring = self.get_limit_paramstring(limit)
    parameters.update({'limit': limit})
    url = project['_meta']['href'] + "/versions" + self._get_parameter_string(parameters)
    headers = self.get_headers()
    headers['Accept'] = 'application/vnd.blackducksoftware.project-detail-4+json'
    response = requests.get(url, headers=headers, verify = not self.config['insecure'])
    jsondata = response.json()
    return jsondata

def get_version_components(self, projectversion, limit=1000):
    paramstring = self.get_limit_paramstring(limit)
    url = projectversion['_meta']['href'] + "/components" + paramstring
    headers = self.get_headers()
    headers['Accept'] = 'application/vnd.blackducksoftware.bill-of-materials-6+json'
    response = requests.get(url, headers=headers, verify = not self.config['insecure'])
    jsondata = response.json()
    return jsondata

def update_project_settings(self, project, new_settings={}):
    url = project['_meta']['href']
    headers = self.get_headers()
    headers['Accept'] = 'application/vnd.blackducksoftware.project-detail-4+json'
    headers['Content-Type'] = 'application/vnd.blackducksoftware.project-detail-4+json'
    response = self.execute_put(url, new_settings, headers)
    return response

def update_project_version_settings(self, project_name, version_name, new_settings={}):
    # Apply any new settings to the given project version
    version = self.get_project_version_by_name(project_name, version_name)

    if version:
        for k,v in new_settings.items():
            if k in PROJECT_VERSION_SETTINGS:
                logger.debug("updating setting {} in version {} with value {}".format(
                    k, version['versionName'], v))
                version[k] = v
            else:
                logger.warn("Setting {} is not in the list of project version settings ({})".format(
                    k, PROJECT_VERSION_SETTINGS))

        url = version['_meta']['href']

        response = self.execute_put(url, version)

        if response.status_code == 200:
            logger.info("Successfully updated version {} with new settings {}".format(
                version['versionName'], new_settings))
        else:
            logger.error("Failed to update version {} with new settings {}; status code: {}".format(
                version['versionName'], new_settings, response.status_code))
    else:
        logger.debug("Did not find a matching project-version in project {}, version name {}".format(
            project_name, version_name))

def get_version_by_id(self, project_id, version_id, limit=100):
    headers = self.get_headers()
    paramstring = self.get_limit_paramstring(limit)
    url = self._get_projects_url() + "/" + project_id + "/versions/" + version_id
    headers['Accept'] = 'application/vnd.blackducksoftware.project-detail-4+json'
    response = requests.get(url, headers=headers, verify = not self.config['insecure'])
    jsondata = response.json()
    return jsondata
    
def compare_project_versions(self, version, compareTo):
    apibase = self.config['baseurl'] + "/api"
    paramstring = "?limit=1000&sortField=component.securityRiskProfile&ascending=false&offset=0"
    cwhat = version['_meta']['href'].replace(apibase, '')
    cto = compareTo['_meta']['href'].replace(apibase, '')
    url = apibase + cwhat + "/compare" + cto + "/components" + paramstring
    headers = self.get_headers()
    response = requests.get(url, headers=headers, verify = not self.config['insecure'])
    jsondata = response.json()
    return jsondata

def get_version_codelocations(self, version, limit=100, offset=0):
    url = self.get_link(version, "codelocations") + self._get_parameter_string({
        'limit': limit,
        'offset': offset})
    custom_headers = {'Content-Type': 'application/vnd.blackducksoftware.scan-4+json'}
    response = self.execute_get(url, custom_headers=custom_headers)
    jsondata = response.json()
    return jsondata

def delete_project_version_by_name(self, project_name, version_name, save_scans=False):
    project = self.get_project_by_name(project_name)
    if project:
        logger.debug("found project {}".format(project))
        project_versions = self.get_project_versions(
            project, 
            parameters={'q':"versionName:{}".format(version_name)}
        )

        project_version_codelocations = None
        if 'totalCount' in project_versions and project_versions['totalCount'] == 1:
            project_version = project_versions['items'][0]
            logger.debug("found the project version: {}".format(project_version))

            delete_scans = not save_scans
            logger.debug("delete_scans was {}".format(delete_scans))

            if delete_scans:
                self.delete_project_version_codelocations(project_version)
            else:
                logger.debug("Delete scans was false, or we did not find any codelocations (scans) in version {} of project {}".format(version_name, project_name))
            # TODO: Check if the project will be "empty" once we delete this version and
            # delete the project accordingly?
            logger.info("Deleting project-version at: {}".format(project_version['_meta']['href']))
            self.execute_delete(project_version['_meta']['href'])
        else:
            logger.debug("Did not find version with name {} in project {}".format(version_name, project_name))
    else:
        logger.debug("Did not find project with name {}".format(project_name))

def delete_project_by_name(self, project_name, save_scans=False, backup_scans=False):
    project = self.get_project_by_name(project_name)
    if project:
        # get project versions
        project_versions = self.get_project_versions(project)
        versions = project_versions.get('items', [])
        logger.debug("Retrieved {} versions for project {}".format(len(versions), project_name))
        
        delete_scans = not save_scans
        logger.debug("delete_scans was {}".format(delete_scans))
            
        if delete_scans:
            # delete all code locations associated with each version
            for version in versions:
                if backup_scans:
                    logger.debug("Backup code locations (aka scans) for version {}".format(version['versionName']))
                    self.download_project_scans(project_name, version['versionName'])
                logger.debug("Deleting code locations (aka scans) for version {}".format(version['versionName']))
                self.delete_project_version_codelocations(version)
                    
        # delete the project itself
        project_url = project['_meta']['href']
        logger.info("Deleting project {}".format(project_name))
        self.execute_delete(project_url)
    else:
        logger.debug("Did not find project with name {}".format(project_name))
        
def delete_project_version_codelocations(self, version):
    version_name = version['versionName']
    try:
        logger.debug("Retrieving code locations (aka scans) for version {}".format(version_name))
        version_code_locations = self.get_version_codelocations(version)
    except:
        logger.error("Failed to get codelocations (aka scans) for version {}".format(version_name), exc_info=True)
        version_code_locations = []
    else:
        version_code_locations = version_code_locations.get('items', []) if version_code_locations else []
    logger.debug("Found {} code locations (aka scans) for version {}".format(len(version_code_locations), version_name))
    code_location_urls = [c['_meta']['href'] for c in version_code_locations]
    for code_location_url in code_location_urls:
        logger.info("Deleting code location at: {}".format(code_location_url))
        self.execute_delete(code_location_url)

def delete_empty_projects(self):
    #get all projects with no mapped code locations and delete them all
    projects = self.get_projects().get('items',[])
    deleted_projects = list()
    for p in projects:
        p_empty = True
        versions = self.get_project_versions(p).get('items', [])
        for v in versions:
            codelocations = self.get_version_codelocations(v)
            if codelocations['totalCount'] != 0:
                p_empty = False
                logger.debug("Found a non-empty version in project {}, skipping...".format(
                    p['name']))
                break
        if p_empty:
            logger.info("Project {} is empty, deleting".format(p['name']))
            self.execute_delete(p['_meta']['href'])
            deleted_projects.append(p['name'])
    return deleted_projects

def delete_empty_versions(self, project):
    # delete versions within a given project if there are no mapped code locations (scans)
    versions = self.get_project_versions(project).get('items', [])
    logger.debug("Deleting empty versions for project {}".format(project['name']))
    deleted_versions = list()
    for v in versions:
        codelocations = self.get_version_codelocations(v).get('items', [])
        if not codelocations:
            logger.info("Deleting empty version {} from project {}".format(
                v['versionName'], project['name']))
            self.execute_delete(v['_meta']['href'])
            deleted_versions.append((project['name'], v['versionName']))
        else:
            logger.debug("Version {} within project {} has scans (i.e. not empty), skipping".format(
                v['versionName'], project['name']))
    return deleted_versions

def delete_all_empty_versions(self):
    # delete versions if there are no mapped code locations (scans) across all projects
    projects = self.get_projects().get('items', [])
    deleted_versions = list()
    logger.info("Deleting empty versions for all {} projects on this server".format(
        len(projects)))
    for p in projects:
        deleted_versions.extend(self.delete_empty_versions(p))
    return deleted_versions

def _find_user_group_url(self, assignable_user_groups, user_group_name):
    for user_group in assignable_user_groups['items']:
        if user_group['name'] == user_group_name:
            return user_group['usergroup']

def _find_user_url(self, assignable_user, user_name):
    for user in assignable_user['items']:
        if user['name'] == user_name:
            return user['user']

def _project_role_urls(self, project_role_names):
    all_project_roles = self.get_project_roles()
    project_role_urls = list()
    for project_role_name in project_role_names:
        for project_role in all_project_roles:
            if project_role_name == project_role['name']:
                project_role_urls.append(project_role['_meta']['href'])
    return project_role_urls

def assign_user_group_to_project(self, project_name, user_group_name, project_roles):
    # Assign the user group to the project using the list of project-role names
    project = self.get_project_by_name(project_name)
    # user_group = self.get_user_group_by_name(user_group_name)

    if project:
        project_url = project['_meta']['href']
        assignable_user_groups_link = self.get_link(project, 'assignable-usergroups')
        if assignable_user_groups_link:
            assignable_user_groups_response = self.execute_get(f"{assignable_user_groups_link}?q=name:{user_group_name}")
            assignable_user_groups = assignable_user_groups_response.json()

            # TODO: What to do if the user group is already assigned to the project, and therefore
            # does not appear in the list of 'assignable' user groups? Should we search the (assigned) user
            # groups and re-apply the project-roles to the assignment?

            user_group_url = self._find_user_group_url(assignable_user_groups, user_group_name)
            if user_group_url:
                headers = self.get_headers()

                # need project role urls to build the POST payload
                project_roles_urls = self._project_role_urls(project_roles)

                # The POST endpoint changes based on whether we found any project-roles to assign
                # Also, due to what appears to be a defect, the Content-Type changes
                if project_roles_urls:
                    url = user_group_url + "/roles"
                    # one dict per project role assignment
                    post_data = [{'role': r, 'scope': project_url} for r in project_roles_urls]
                    # I found I had to use this Content-Type (application/json resulted in 412)
                    # ref: https://jira.dc1.lan/browse/HUB-18417
                    headers['Content-Type'] = 'application/vnd.blackducksoftware.internal-1+json'
                else:
                    url = project_url + "/usergroups"
                    # Assigning a group with no project-roles
                    post_data = {"group": user_group_url}
                    headers['Content-Type'] = 'application/json'

                response = requests.post(
                    url, 
                    headers=headers, 
                    data=json.dumps(post_data), 
                    verify = not self.config['insecure'])
                return response
            else:
                assignable_groups = [u['name'] for u in assignable_user_groups['items']]
                logger.warning("The user group {} was not found in the assignable user groups ({}) for this project {}. Is the group already assigned to this project?".format(
                    user_group_name, assignable_groups, project_name))
        else:
            logger.warning("This project {} has no assignable user groups".format(project_name))
    else:
        logger.warning("Did not find a project by the name {}".format(project_name))

def delete_user_group_from_project(self, project_name, user_group_name):
    project = self.get_project_by_name(project_name)
    
    if project:
        project_url = project['_meta']['href']
        
        user_group = self.get_user_group_by_name(user_group_name)
        if user_group:
            user_group_url = user_group['_meta']['href']
            user_group_id = user_group_url.rsplit('/', 1)[-1]
        
            project_user_group_url = f"{project_url}/usergroups/{user_group_id}"
            self.execute_delete(project_user_group_url)    

def assign_user_to_project(self, user_name, project_name, project_roles, limit=1000):
    # Assign users to projects
    project = self.get_project_by_name(project_name)

    if project:
        project_url = project['_meta']['href']
        assignable_users_link = self.get_link(project, 'assignable-users')
        paramstring = self.get_limit_paramstring(limit)
        url = assignable_users_link + paramstring
        logger.debug("GET {}".format(url))
        if assignable_users_link:
            assignable_users_response = self.execute_get(url)
            assignable_users = assignable_users_response.json()

            # TODO: What to do if the user is already assigned to the project, and therefore
            # does not appear in the list of 'assignable' user? Should we search the (assigned) user
            # and re-apply the project-roles to the assignment?

            user_url = self._find_user_url(assignable_users, user_name)
            if user_url:
                headers = self.get_headers()

                # need project role urls to build the POST payload
                project_roles_urls = self._project_role_urls(project_roles)

                # The POST endpoint changes based on whether we found any project-roles to assign
                # Also, due to what appears to be a defect, the Content-Type changes
                if project_roles_urls:
                    url = user_url + "/roles"
                    # one dict per project role assignment
                    post_data = [{'role': r, 'scope': project_url} for r in project_roles_urls]
                    # I found I had to use this Content-Type (application/json resulted in 412)
                    # ref: https://jira.dc1.lan/browse/HUB-18417
                    headers['Content-Type'] = 'application/vnd.blackducksoftware.internal-1+json'
                else:
                    url = project_url + "/users"
                    # Assigning a user with no project-roles
                    post_data = {"user": user_url}
                    headers['Content-Type'] = 'application/json'

                response = requests.post(
                    url,
                    headers=headers,
                    data=json.dumps(post_data),
                    verify=not self.config['insecure'])
                return response
            else:
                assignable_username = [u['name'] for u in assignable_users['items']]
                logger.warning(
                    "The user {} was not found in the assignable user ({}) for this project {}. Is the user already assigned to this project?".format(
                        user_name, assignable_username, project_name))
        else:
            logger.warning("This project {} has no assignable users".format(project_name))
    else:
        logger.warning("Did not find a project by the name {}".format(project_name))

def assign_project_application_id(self, project_name, application_id, overwrite=False):
    logger.debug("Assigning application_id {} to project_name {}, overwrite={}".format(
        application_id, project_name, overwrite))

    existing_application_id, application_id_url = self.get_project_application_id(project_name)

    if existing_application_id:
        if overwrite:
            logger.debug("Found an existing application id {} for project {} and overwrite was True. Updating it to {}".format(
                existing_application_id, project_name, application_id))
            return self.update_project_application_id(project_name, application_id)
        else:
            logger.debug("Found an existing application id {} for project {} and overwrite was False so not updating it".format(
                existing_application_id, project_name))
    else:
        logger.debug("No application id exists for project {}, assigning {} to it".format(
            project_name, application_id))
        project = self.get_project_by_name(project_name)
        if project:
            project_mappings_url = self.get_link(project, "project-mappings")
            if project_mappings_url:
                post_data = {"applicationId": application_id}
                response = self.execute_post(project_mappings_url, data=post_data)
                return response
            else:
                logger.warning("Did not find project-mappings URL for project {}".format(project))
        else:
            logger.warning("Did not find project by name {}".format(project_name))

def update_project_application_id(self, project_name, new_application_id):
    application_id, application_id_url = self.get_project_application_id(project_name)

    if application_id and application_id_url:
        put_data = {
            "applicationId": new_application_id,
            "_meta": {
                "allow": [
                    "DELETE",
                    "GET",
                    "PUT"
                ],
                "href": application_id_url,
                "links": []
            }
        }
        response = self.execute_put(application_id_url, data=put_data)
        return response
    else:
        logger.debug("Did not find application id for project name {}".format(project_name))

def delete_application_id(self, project_name):
    application_id, application_id_url = self.get_project_application_id(project_name)

    if application_id_url:
        self.execute_delete(application_id_url)

def get_project_application_id(self, project_name):
    project_mapping_info = self.get_project_info(project_name, 'project-mappings')
    if project_mapping_info and 'items' in project_mapping_info:
        for project_mapping in project_mapping_info['items']:
            if 'applicationId' in project_mapping:
                application_id = project_mapping['applicationId']
                application_id_url = project_mapping['_meta']['href']

                return (application_id, application_id_url)
        logger.debug("Did not find any project-mappings with 'applicationId' in them")
        return (None, None)
    else:
        logger.debug("did not find any project-mappings for project {}".format(project_name))
        return (None, None)

def get_project_info(self, project_name, link_name):
    project = self.get_project_by_name(project_name)
    link = self.get_link(project, link_name)
    if link:
        response = self.execute_get(link)
        return response.json()
    else:
        return {} # nada

def get_project_roles(self):
    all_project_roles = self.get_roles(parameters={"filter":"scope:project"})
    return all_project_roles['items']

def get_version_scan_info(self, version_obj):
    url = self.get_link(version_obj, "codelocations")
    custom_headers = {'Accept': 'application/vnd.blackducksoftware.project-detail-5+json'}
    response = self.execute_get(url, custom_headers=custom_headers)
    code_locations = response.json().get('items', [])
    if code_locations:
        scan_info = {
            'most_recent_scan': max([cl['updatedAt'] for cl in code_locations]),
            'oldest_scan': min([cl['createdAt'] for cl in code_locations]),
            'number_scans': len(code_locations)
        }
    else:
        scan_info = {
            'most_recent_scan': None,
            'oldest_scan': None,
            'number_scans': None
        }
    return scan_info



import logging
import requests
import json
from operator import itemgetter
import urllib.parse

logger = logging.getLogger(__name__)

def _get_vulnerabilities_url(self):
    return self.config['baseurl'] + '/api/vulnerabilities'

def get_vulnerabilities(self, vulnerability, parameters={}):
    url = self._get_vulnerabilities_url() + "/{}".format(vulnerability) + self._get_parameter_string(parameters)
    headers = {'Accept': 'application/vnd.blackducksoftware.vulnerability-4+json'}
    response = self.execute_get(url, custom_headers=headers)
    return response.json()

def get_vulnerability_affected_projects(self, vulnerability):
    url = self._get_vulnerabilities_url() + "/{}/affected-projects".format(vulnerability) 
    custom_headers = {'Accept': 'application/vnd.blackducksoftware.vulnerability-4+json'}
    response = self.execute_get(url, custom_headers=custom_headers)
    return response.json()

# TODO: Refactor this, i.e. use get_link method?
def get_vulnerable_bom_components(self, version_obj, limit=9999):
    url = "{}/vulnerable-bom-components".format(version_obj['_meta']['href'])
    custom_headers = {'Accept': 'application/vnd.blackducksoftware.bill-of-materials-6+json'}
    param_string = self._get_parameter_string({'limit': limit})
    url = "{}{}".format(url, param_string)
    response = self.execute_get(url, custom_headers=custom_headers)
    return response.json()

# TODO: Remove or refactor this
def get_component_remediation(self, bom_component):
    url = "{}/remediating".format(bom_component['componentVersion'])
    logger.debug("Url for getting remediation info is : {}".format(url))
    response = self.execute_get(url)
    return response.json()



import logging
import requests
import json
from operator import itemgetter
import urllib.parse

logger = logging.getLogger(__name__)

def _get_cf_url(self):
    return self.get_apibase() + "/custom-fields/objects"

def supported_cf_object_types(self):
    '''Get the types and cache them since they are static (on a per-release basis)'''
    if not hasattr(self, "_cf_object_types"):
        logger.debug("retrieving object types")
        self._cf_object_types = [cfo['name'] for cfo in self.get_cf_objects().get('items', [])]
    return self._cf_object_types

def get_cf_objects(self):
    '''Get CF objects and cache them since these are static (on a per-release basis)'''
    url = self._get_cf_url()
    if not hasattr(self, "_cf_objects"):
        logger.debug("retrieving objects")
        response = self.execute_get(url)
        self._cf_objects = response.json()
    return self._cf_objects

def _get_cf_object_url(self, object_name):
    for cf_object in self.get_cf_objects().get('items', []):
        if cf_object['name'].lower() == object_name.lower():
            return cf_object['_meta']['href']

def get_cf_object(self, object_name):
    assert object_name in self.supported_cf_object_types(), "Object name {} not one of the supported types ({})".format(object_name, self.supported_cf_object_types())

    object_url = self._get_cf_object_url(object_name)
    response = self.execute_get(object_url)
    return response.json()

def _get_cf_obj_rel_path(self, object_name):
    return object_name.lower().replace(" ", "-")

def create_cf(self, object_name, field_type, description, label, position, active=True, initial_options=[]):
    '''
        Create a custom field for the given object type (e.g. "Project", "Project Version") using the field_type and other parameters.

        Initial options are needed for field types like multi-select where the multiple values to choose from must also be provided.

        initial_options = [{"label":"val1", "position":0}, {"label":"val2", "position":1}]
    '''
    assert isinstance(position, int) and position >= 0, "position must be an integer that is greater than or equal to 0"
    assert field_type in ["BOOLEAN", "DATE", "DROPDOWN", "MULTISELECT", "RADIO", "TEXT", "TEXTAREA"]

    types_using_initial_options = ["DROPDOWN", "MULTISELECT", "RADIO"]

    post_url = self._get_cf_object_url(object_name) + "/fields"
    cf_object = self._get_cf_obj_rel_path(object_name)
    cf_request = {
        "active": active,
        "description": description,
        "label": label,
        "position": position,
        "type": field_type,
    }
    if field_type in types_using_initial_options and initial_options:
        cf_request.update({"initialOptions": initial_options})
    response = self.execute_post(post_url, data=cf_request)
    return response

def delete_cf(self, object_name, field_id):
    '''Delete a custom field from a given object type, e.g. Project, Project Version, Component, etc

    WARNING: Deleting a custom field is irreversiable. Any data in the custom fields could be lost so use with caution.
    '''
    assert object_name in self.supported_cf_object_types(), "You must supply a supported object name that is in {}".format(self.supported_cf_object_types())

    delete_url = self._get_cf_object_url(object_name) + "/fields/{}".format(field_id)
    return self.execute_delete(delete_url)

def get_custom_fields(self, object_name):
    '''Get the custom field (definition) for a given object type, e.g. Project, Project Version, Component, etc
    '''
    assert object_name in self.supported_cf_object_types(), "You must supply a supported object name that is in {}".format(self.supported_cf_object_types())

    url = self._get_cf_object_url(object_name) + "/fields"

    response = self.execute_get(url)        
    return response.json()

def get_cf_values(self, obj):
    '''Get all of the custom fields from an object such as a Project, Project Version, Component, etc

    The obj is expected to be the JSON document for a project, project-version, component, etc
    '''
    url = self.get_link(obj, "custom-fields")
    response = self.execute_get(url)
    return response.json()

def get_cf_value(self, obj, field_id):
    '''Get a custom field value from an object such as a Project, Project Version, Component, etc

    The obj is expected to be the JSON document for a project, project-version, component, etc
    '''
    url = self.get_link(obj, "custom-fields") + "/{}".format(field_id)
    response = self.execute_get(url)
    return response.json()

def put_cf_value(self, cf_url, new_cf_obj):
    '''new_cf_obj is expected to be a modified custom field value object with the values updated accordingly, e.g.
    call get_cf_value, modify the object, and then call put_cf_value
    '''
    return self.execute_put(cf_url, new_cf_obj)



#!/usr/bin/env python

import copy
import json
import os
import pytest
import re
from urllib.parse import urlparse

from blackduck.HubRestApi import HubInstance
from unittest.mock import patch, MagicMock, mock_open


fake_hub_host = "https://my-hub-host"


def return_auth_token(api_token):
    if api_token:
        return ("the-token", "the-csrf-token")
    else:
        return ("the-token", None)

invalid_bearer_token="anInvalidTokenValue"
invalid_csrf_token="anInvalidCSRFTokenValue"
made_up_api_token="theMadeUpAPIToken"

def setup_function(function):
    # Remove .restconfig file before running any test
    try:
        os.remove(HubInstance.configfile)
    except OSError:
        pass

def teardown_function(function):
    # Remove .restconfig file after running any test
    try:
        os.remove(HubInstance.configfile)
    except OSError:
        pass
        
@pytest.fixture()
def mock_hub_instance(requests_mock):
    requests_mock.post(
        "{}/j_spring_security_check".format(fake_hub_host), 
        headers={"Set-Cookie": 'AUTHORIZATION_BEARER={}; Path=/; secure; Secure; HttpOnly'.format(invalid_bearer_token)}
    )
    requests_mock.get(
        "{}/api/current-version".format(fake_hub_host),
        json = {
            "version": "2018.11.0",
            "_meta": {
                "allow": [
                    "GET"
                ],
                "href": "{}/api/current-version".format(fake_hub_host)
            }
        }
    )
    yield HubInstance(fake_hub_host, "a_username", "a_password")

@pytest.fixture()
def mock_hub_instance_using_api_token(requests_mock):
    requests_mock.post(
        "https://my-hub-host/api/tokens/authenticate", 
        content = json.dumps({'bearerToken': invalid_bearer_token}).encode('utf-8'),
        headers={
                'X-CSRF-TOKEN': invalid_csrf_token, 
                'Content-Type': 'application/json'
            }
    )
    requests_mock.get(
        "{}/api/current-version".format(fake_hub_host),
        json = {
            "version": "2018.11.0",
            "_meta": {
                "allow": [
                    "GET"
                ],
                "href": "{}/api/current-version".format(fake_hub_host)
            }
        }
    )

    yield HubInstance(fake_hub_host, api_token=made_up_api_token)

@pytest.fixture()
def policy_info_json(shared_datadir):
    yield json.load((shared_datadir / "policies.json").open())

@pytest.fixture()
def a_test_policy(policy_info_json):
    test_policy = policy_info_json['items'][0]
    yield test_policy

@pytest.fixture()
def a_test_policy_for_create_or_update(a_test_policy):
        # a_policy_for_creating_or_updating = dict(
        #     (attr, test_policy[attr]) for attr in 
        #     ['name', 'description', 'enabled', 'overridable', 'expression', 'severity'] if attr in test_policy)
        # yield a_policy_for_creating_or_updating
        yield a_test_policy

@pytest.fixture()
def test_vulnerability_info(requests_mock, shared_datadir):
    yield json.loads((shared_datadir / "sample-vulnerability.json").read_text())

def test_get_major_version(requests_mock):
    requests_mock.post(
        "{}/j_spring_security_check".format(fake_hub_host), 
        headers={"Set-Cookie": 'AUTHORIZATION_BEARER={}; Path=/; secure; Secure; HttpOnly'.format(invalid_bearer_token)}
    )
    for version in ["2018.11.0", "5.0.2", "4.8.3", "3.7.2"]:
        expected = version.split(".")[0]
        requests_mock.get(
            "{}/api/current-version".format(fake_hub_host),
            json = {
                "version": "{}".format(version),
                "_meta": {
                    "allow": [
                        "GET"
                    ],
                    "href": "{}/api/current-version".format(fake_hub_host)
                }
            }
        )
        hub = HubInstance(fake_hub_host, "a_username", "a_password")
        assert hub.bd_major_version == expected

def test_get_headers(mock_hub_instance):
    # somewhat contrived, but it does execute all the paths
    # TODO: better way to test this one?
    #
    the_api_token = "fake-api-token"
    the_csrf_token = "fake-csrf-token"
    the_token = "fake-bearer-token"
    mock_hub_instance.config['api_token'] = the_api_token
    mock_hub_instance.csrf_token = the_csrf_token
    mock_hub_instance.token = the_token

    assert mock_hub_instance.get_headers() == {
                'X-CSRF-TOKEN': the_csrf_token, 
                'Authorization': "Bearer {}".format(the_token),
                'Content-Type': 'application/json',
                'Accept': 'application/json'}

    del mock_hub_instance.config['api_token']
    for bd_major_version in ["2018", "5", "4", "3"]:
        if bd_major_version == "3":
            expected =  {"Cookie": mock_hub_instance.cookie}
        else:
            expected =  {"Authorization":"Bearer " + mock_hub_instance.token}

        mock_hub_instance.bd_major_version = bd_major_version
        assert mock_hub_instance.get_headers() == expected

def test_get_policy_url(mock_hub_instance):
    assert mock_hub_instance._get_policy_url() == fake_hub_host + "/api/policy-rules"

def test_get_parameter_string(mock_hub_instance):
    assert mock_hub_instance._get_parameter_string({"limit":"100"}) == "?limit=100"
    assert mock_hub_instance._get_parameter_string({"limit":"100", "q":"name:my-name"}) == "?limit=100&q=name%3Amy-name"
    assert mock_hub_instance._get_parameter_string({"limit":"100", "sort":"updatedAt"}) == "?limit=100&sort=updatedAt"

def test_hub_instance_username_password_for_auth(mock_hub_instance):
    assert mock_hub_instance.get_headers() == {"Authorization":"Bearer {}".format(invalid_bearer_token)}

    assert 'api_token' not in mock_hub_instance.config
    assert 'baseurl' in mock_hub_instance.config
    assert 'username' in mock_hub_instance.config
    assert 'password' in mock_hub_instance.config

def test_hub_instance_api_token_for_auth(mock_hub_instance_using_api_token):
    assert mock_hub_instance_using_api_token.get_headers() == {
                'X-CSRF-TOKEN': invalid_csrf_token, 
                'Authorization': 'Bearer {}'.format(invalid_bearer_token), 
                'Content-Type': 'application/json',
                'Accept': 'application/json'}

    assert 'api_token' in mock_hub_instance_using_api_token.config
    assert 'baseurl' in mock_hub_instance_using_api_token.config
    assert 'username' not in mock_hub_instance_using_api_token.config
    assert 'password' not in mock_hub_instance_using_api_token.config

def test_hub_instance_with_write_config(requests_mock):
    requests_mock.post(
        "https://my-hub-host/j_spring_security_check", 
        headers={"Set-Cookie": 'AUTHORIZATION_BEARER={}; Path=/; secure; Secure; HttpOnly'.format(invalid_bearer_token)}
    )
    requests_mock.get(
        "{}/api/current-version".format(fake_hub_host),
        json = {
            "version": "2018.11.0",
            "_meta": {
                "allow": [
                    "GET"
                ],
                "href": "{}/api/current-version".format(fake_hub_host)
            }
        }
    )
    
    with patch("builtins.open", new_callable=mock_open()) as m:
        with patch('json.dump') as m_json:
            hub = HubInstance(fake_hub_host, "a_username", "a_password")

            m.assert_called_with('.restconfig.json', 'w')
            assert m_json.called

def test_hub_instance_with_write_config_false(requests_mock):
    requests_mock.post(
        "https://my-hub-host/j_spring_security_check", 
        headers={"Set-Cookie": 'AUTHORIZATION_BEARER={}; Path=/; secure; Secure; HttpOnly'.format(invalid_bearer_token)}
    )
    requests_mock.get(
        "{}/api/current-version".format(fake_hub_host),
        json = {
            "version": "2018.11.0",
            "_meta": {
                "allow": [
                    "GET"
                ],
                "href": "{}/api/current-version".format(fake_hub_host)
            }
        }
    )

    with patch.object(HubInstance, "write_config") as mock_write_config:
        hub = HubInstance(fake_hub_host, "a_username", "a_password", write_config_flag=False)

        assert not mock_write_config.called

def test_get_policy_by_id(requests_mock, mock_hub_instance, a_test_policy):
    requests_mock.get(fake_hub_host + "/api/policy-rules/00000000-0000-0000-0000-000000000001", json=a_test_policy)
    policy = mock_hub_instance.get_policy_by_id("00000000-0000-0000-0000-000000000001")
    for key in a_test_policy.keys():
        assert policy[key] == a_test_policy[key]

def test_get_policy_by_url(requests_mock, mock_hub_instance, a_test_policy):
    requests_mock.get(fake_hub_host + "/api/policy-rules/00000000-0000-0000-0000-000000000001", json=a_test_policy)
    policy = mock_hub_instance.get_policy_by_url(mock_hub_instance._get_policy_url() + "/00000000-0000-0000-0000-000000000001")
    for key in a_test_policy.keys():
        assert policy[key] == a_test_policy[key]

def test_update_policy_by_id(requests_mock, mock_hub_instance, a_test_policy, a_test_policy_for_create_or_update):
    policy_id = a_test_policy['_meta']['href'].split("/")[-1]

    requests_mock.put(fake_hub_host + "/api/policy-rules/" + policy_id,
        json=a_test_policy
    )
    response = mock_hub_instance.update_policy_by_id(policy_id, a_test_policy_for_create_or_update)
    assert response.status_code == 200
    assert response.json() == a_test_policy

def test_update_policy_by_url(requests_mock, mock_hub_instance, a_test_policy, a_test_policy_for_create_or_update):
    policy_id = a_test_policy['_meta']['href'].split("/")[-1]
    policy_url = mock_hub_instance._get_policy_url() + "/" + policy_id

    requests_mock.put(fake_hub_host + "/api/policy-rules/" + policy_id,
        json=a_test_policy
    )
    response = mock_hub_instance.update_policy_by_url(policy_url, a_test_policy_for_create_or_update)
    assert response.status_code == 200
    assert response.json() == a_test_policy

def test_create_policy(requests_mock, mock_hub_instance, a_test_policy, a_test_policy_for_create_or_update):
    requests_mock.post(fake_hub_host + "/api/policy-rules", headers={"location": a_test_policy['_meta']['href']}, status_code=201)
    # print(json.dumps(a_test_policy_for_create_or_update))
    new_policy_url = mock_hub_instance.create_policy(a_test_policy_for_create_or_update)
    assert new_policy_url == a_test_policy['_meta']['href']

def test_delete_policy_by_id(requests_mock, mock_hub_instance, a_test_policy):
    policy_id = a_test_policy['_meta']['href'].split("/")[-1]

    requests_mock.delete(fake_hub_host + "/api/policy-rules/" + a_test_policy['_meta']['href'].split("/")[-1], status_code=204)
    response = mock_hub_instance.delete_policy_by_id(policy_id)
    assert response.status_code == 204

def test_delete_policy_by_url(requests_mock, mock_hub_instance, a_test_policy):
    policy_url = mock_hub_instance._get_policy_url() + "/" + a_test_policy['_meta']['href'].split("/")[-1]

    requests_mock.delete(policy_url, status_code=204)
    response = mock_hub_instance.delete_policy_by_url(policy_url)
    assert response.status_code == 204

def test_get_vulnerability(requests_mock, mock_hub_instance, test_vulnerability_info):
    vulnerability_url = mock_hub_instance._get_vulnerabilities_url() + "/{}".format(test_vulnerability_info['vulnerabilityName'])

    requests_mock.get(vulnerability_url, json=test_vulnerability_info)
    response_json = mock_hub_instance.get_vulnerabilities(test_vulnerability_info['vulnerabilityName'])

    assert response_json == test_vulnerability_info

def test_get_projects_with_limit(requests_mock, mock_hub_instance, shared_datadir):
    url = mock_hub_instance.get_urlbase() + "/api/projects?limit=20"
    json_data = json.load((shared_datadir / 'sample-projects.json').open())
    requests_mock.get(url, json=json_data)
    projects = mock_hub_instance.get_projects(limit=20)

    assert json_data == projects
    assert 'totalCount' in projects
    assert projects['totalCount'] == 18

def test_get_projects_with_name_query(requests_mock, mock_hub_instance, shared_datadir):
    url = mock_hub_instance.get_urlbase() + "/api/projects?q=name:accelerator-initializer-ui&limit=100"
    json_data = json.load((shared_datadir / 'sample-projects-using-name-query.json').open())
    requests_mock.get(url, json=json_data)
    projects = mock_hub_instance.get_projects(parameters={'q':"name:accelerator-initializer-ui"})

    assert json_data == projects
    assert 'totalCount' in projects
    assert projects['totalCount'] == 1

def test_get_project_versions(requests_mock, mock_hub_instance, shared_datadir):
    baseurl = mock_hub_instance.get_urlbase()
    url = baseurl + "/api/projects/65f272df-3a2a-4022-8811-a57e05e82f52/versions?limit=100"
    json_data = json.load((shared_datadir / 'sample-project-versions.json').open())
    project_json_data = json.load((shared_datadir / 'sample-project.json').open())
    # replace project URL with the right one to agree with our mocked URL above
    project_json_data['_meta']['href'] = re.sub("https://.*/api", "{}/api".format(baseurl), project_json_data['_meta']['href'])
    requests_mock.get(url, json=json_data)
    versions = mock_hub_instance.get_project_versions(project_json_data)

    assert 'totalCount' in versions
    assert versions['totalCount'] == 1

def test_get_project_versions_with_parameters(requests_mock, mock_hub_instance, shared_datadir):
    baseurl = mock_hub_instance.get_urlbase()
    url = baseurl + "/api/projects/65f272df-3a2a-4022-8811-a57e05e82f52/versions?limit=100&q=versionName:1.0"
    json_data = json.load((shared_datadir / 'sample-project-versions.json').open())
    project_json_data = json.load((shared_datadir / 'sample-project.json').open())
    # replace project URL with the right one to agree with our mocked URL above
    project_json_data['_meta']['href'] = re.sub("https://.*/api", "{}/api".format(baseurl), project_json_data['_meta']['href'])
    requests_mock.get(url, json=json_data)
    versions = mock_hub_instance.get_project_versions(project_json_data, parameters={'q':'versionName:1.0'})

    assert 'totalCount' in versions
    assert versions['totalCount'] == 1
    assert 'items' in versions

def test_delete_project_version_by_name():
    # TODO: Write this test
    pass


def test_get_users(requests_mock, mock_hub_instance, shared_datadir):
    baseurl = mock_hub_instance.get_urlbase()
    url = baseurl + "/api/users"
    user_json_data = json.load((shared_datadir / "users.json").open())
    requests_mock.get(url, json=user_json_data)
    users = mock_hub_instance.get_users()

    assert 'totalCount' in users
    assert users['totalCount'] == 1 # cause there was one user in the sample data collected
    assert 'items' in users

def test_create_user(requests_mock, mock_hub_instance):
    pass

def test_get_user_by_id(requests_mock, mock_hub_instance):
    pass

def test_get_user_by_url(requests_mock, mock_hub_instance):
    pass

def test_update_user_by_id(requests_mock, mock_hub_instance):
    pass

def test_update_user_by_url(requests_mock, mock_hub_instance):
    pass

def test_delete_user_by_id(requests_mock, mock_hub_instance):
    pass

def test_delete_user_by_url(requests_mock, mock_hub_instance):
    pass
    
def test_get_project_by_name(requests_mock, mock_hub_instance, shared_datadir):
    url = mock_hub_instance.get_urlbase() + "/api/projects"
    projects_json = json.load((shared_datadir / "sample-projects.json").open())
    project_name = "accelerator-initializer-ui"
    requests_mock.get(url, json=projects_json)

    project = mock_hub_instance.get_project_by_name(project_name)

    assert project['name'] == project_name

def test_get_version_by_name(requests_mock, mock_hub_instance, shared_datadir):
    mock_hub_instance.get_project_versions = MagicMock(return_value=json.load((shared_datadir / "sample-project-versions.json").open()))

    mock_project_obj = MagicMock()
    version_name = "1.0" # a version that exists in sample-project-versions.json
    version = mock_hub_instance.get_version_by_name(mock_project_obj, version_name)

    assert version['versionName'] == version_name

def test_create_version_reports(requests_mock, mock_hub_instance):
    pass

def test_create_version_notices_report(requests_mock, mock_hub_instance):
    pass

@pytest.fixture()
def sample_bom_component_json(shared_datadir):
    yield json.load((shared_datadir / "sample-bom-component.json").open())

@pytest.fixture()
def no_roles_user(shared_datadir):
    yield json.load((shared_datadir / 'no-roles-user.json').open())

@pytest.fixture()
def no_roles_roles(shared_datadir):
    yield json.load((shared_datadir / 'no-roles-roles.json').open())

@pytest.fixture()
def sysadmin_user(shared_datadir):
    yield json.load((shared_datadir / 'sysadmin-user.json').open())

@pytest.fixture()
def sysadmin_roles(shared_datadir):
    yield json.load((shared_datadir / 'sysadmin-roles.json').open())

def test_user_has_role(no_roles_user, no_roles_roles, sysadmin_user, sysadmin_roles, mock_hub_instance):
    test_roles = [
        'License Manager', 
        'System Administrator', 
        'Policy Manager', 
        'Project Viewer', 
        'Global Code Scanner', 
        'Project Manager']
    
    mock_hub_instance.get_roles_for_user_or_group = MagicMock()
    mock_hub_instance.get_roles_for_user_or_group.return_value = no_roles_roles

    for test_role in test_roles:
        assert mock_hub_instance.user_has_role(no_roles_user, test_role) == False

    mock_hub_instance.get_roles_for_user_or_group.return_value = sysadmin_roles

    for test_role in test_roles:
        assert mock_hub_instance.user_has_role(sysadmin_user, test_role) == True

def test_get_link(mock_hub_instance):
    a_url = 'http://a-url'
    link_name = 'a_link_name'
    bd_rest_obj = {'_meta':{'links': [{'rel': link_name, 'href': a_url}]}}

    assert mock_hub_instance.get_link(bd_rest_obj, link_name) == a_url

def test_get_link_returns_None_for_invalid_bd_rest_object(mock_hub_instance):
    bd_rest_obj = {}

    assert mock_hub_instance.get_link(bd_rest_obj, 'a_link_name') == None

def test_validated_json_with_a_dictionary(mock_hub_instance):
    validated_json = mock_hub_instance._validated_json_data({'key':'value'})
    assert isinstance(validated_json, str)
    assert validated_json == '{"key": "value"}'

def test_validated_json_with_a_list(mock_hub_instance):
    validated_json = mock_hub_instance._validated_json_data(['item1', 'item2'])
    assert isinstance(validated_json, str)
    assert validated_json == '["item1", "item2"]'

def test_validated_json_with_a_json_str(mock_hub_instance):
    validated_json = mock_hub_instance._validated_json_data('{"key": "value"}')

    assert validated_json == '{"key": "value"}'

def test_validated_json_fails_with_invalid_json_str(mock_hub_instance):
    from json import JSONDecodeError

    with pytest.raises(JSONDecodeError) as e_info:
        validated_json = mock_hub_instance._validated_json_data('invalid json')

@pytest.fixture()
def code_locations(requests_mock, shared_datadir):
    data = json.load((shared_datadir / 'code_locations.json').open())
    requests_mock.get(
        "{}/api/codelocations?limit=100".format(fake_hub_host),
        json = data
    )
    yield data

def test_get_codelocations_all(mock_hub_instance, code_locations):
    code_locs = mock_hub_instance.get_codelocations()

    assert code_locs == code_locations

def test_get_codelocations_unmapped(mock_hub_instance, code_locations):
    code_locs = mock_hub_instance.get_codelocations(unmapped=True)

    assert code_locs != code_locations

    expected_data = copy.deepcopy(code_locations)
    expected_data['items'] = [cl for cl in code_locations['items'] if 'mappedProjectVersion' not in cl]
    expected_data['totalCount'] = len(expected_data['items'])

    assert code_locs == expected_data





import logging
import requests
import json
from operator import itemgetter
import urllib.parse

logger = logging.getLogger(__name__)

def find_component_info_for_protex_component(self, protex_component_id, protex_component_release_id):
    '''Will return the Hub component corresponding to the protex_component_id, and if a release (version) id
    is given, the response will also include the component-version. Returns an empty list if there were
    no components found.
    '''
    url = self.config['baseurl'] + "/api/components"
    if protex_component_release_id:
        query = "?q=bdsuite:{}%23{}&limit=9999".format(protex_component_id, protex_component_release_id)
    else:
        query = "?q=bdsuite:{}&limit=9999".format(protex_component_id)
    with_query = url + query
    logger.debug("Finding the Hub componet for Protex component id {}, release id {} using query/url {}".format(
        protex_component_id, protex_component_release_id, with_query))
    response = self.execute_get(with_query)
    logger.debug("query results in status code {}, json data: {}".format(response.status_code, response.json()))
    # TODO: Error checking and retry? For now, as POC just assuming it worked
    component_list_d = response.json()
    return response.json()

def _get_components_url(self):
    return self.get_urlbase() + "/api/components"

def get_components(self, limit=100, parameters={}):
    if limit:
        parameters.update({'limit':limit})
    #
    # I was only able to GET components when using this internal media type which is how the GUI works
    #       July 19, 2019 Glenn Snyder
    #
    custom_headers = {'Accept':'application/vnd.blackducksoftware.internal-1+json'}
    url = self._get_components_url() + self._get_parameter_string(parameters)
    response = self.execute_get(url, custom_headers=custom_headers)
    return response.json()

def search_components(self, search_str_or_query, limit=100, parameters={}):
    if limit:
        parameters.update({'limit':limit})
    if search_str_or_query.startswith("q="):
        # allow caller to override original behavior with their own query
        query = search_str_or_query
    else:
        # maintain original, somewhat flawed behavior
        query = "q=name:{}".format(search_str_or_query)
    parm_str = self._get_parameter_string(parameters)
    url = self.get_apibase() + "/search/components{}&{}".format(parm_str, query)
    response = self.execute_get(url)
    return response.json()
    
def get_component_by_id(self, component_id):
    url = self.config['baseurl'] + "/api/components/{}".format(component_id)
    return self.get_component_by_url(url)

def get_component_by_url(self, component_url):
    headers = self.get_headers()
    response = self.execute_get(component_url)
    jsondata = response.json()
    return jsondata

def update_component_by_id(self, component_id, update_json):
    url = self.config["baseurl"] + "/api/components/{}".format(component_id)
    return self.update_component_by_url(url, update_json)

def update_component_by_url(self, component_url, update_json):
    return self.execute_put(component_url, update_json)



'''

Created on Dec 22, 2020
@author: ar-calder

'''
import logging
from json import JSONDecodeError
from .Utils import pfmt

logger = logging.getLogger(__name__)

class CreateFailedAlreadyExists(Exception):
    pass

class CreateFailedUnknown(Exception):
    pass

class InvalidVersionPhase(Exception):
    pass

class UnknownVersion(Exception):
    pass

class UnsupportedBDVersion(Exception):
    # Some operations require specific versions of BD
    pass

class EndpointNotFound(Exception):
    pass

class UnacceptableContentType(Exception):
    pass

def http_exception_handler(self, response, name):
    error_codes = {
        404: EndpointNotFound,
        406: UnacceptableContentType
    }

    try:
        content = pfmt(response.json())
    except JSONDecodeError:
        content = response.text

    error = error_codes.get(response.status_code)
    if error:
        raise error(f"(status code {response.status_code}) {name}: {content}")
    raise NotImplementedError(f"No handler for status code: {response.status_code}")



import logging
import requests
import json
from operator import itemgetter
import urllib.parse

logger = logging.getLogger(__name__)

def _get_policy_url(self):
    return self.config['baseurl'] + "/api/policy-rules"

def get_policies(self, parameters={}):
    url = self._get_policy_url() + self._get_parameter_string(parameters)
    headers = {'Accept': 'application/json'}
    response = self.execute_get(url, custom_headers=headers)
    return response.json()

def create_policy(self, policy_json):
    url = self._get_policy_url()
    location = self._create(url, policy_json)
    return location

def get_policy_by_id(self, policy_id):
    url = self._get_policy_url() + "/{}".format(policy_id)
    return self.get_policy_by_url(url)

def get_policy_by_url(self, policy_url):
    headers = {'Accept': 'application/vnd.blackducksoftware.policy-4+json'}
    response = self.execute_get(policy_url, custom_headers=headers)
    jsondata = response.json()
    return jsondata

def update_policy_by_id(self, policy_id, update_json):
    url = self._get_policy_url() + "/{}".format(policy_id)
    return self.update_policy_by_url(url, update_json)

def update_policy_by_url(self, policy_url, update_json):
    return self.execute_put(policy_url, update_json)

def delete_policy_by_id(self, policy_id):
    url = self._get_policy_url() + "/{}".format(policy_id)
    return self.delete_policy_by_url(url)

def delete_policy_by_url(self, policy_url):
    return self.execute_delete(policy_url)



import logging
import requests
import json
from operator import itemgetter
import urllib.parse

from .Utils import object_id

logger = logging.getLogger(__name__)

valid_categories = ['VERSION','CODE_LOCATIONS','COMPONENTS','SECURITY','FILES', 'ATTACHMENTS', 'CRYPTO_ALGORITHMS', 'PROJECT_VERSION_CUSTOM_FIELDS', 'BOM_COMPONENT_CUSTOM_FIELDS', 'LICENSE_TERM_FULFILLMENT', 'UPGRADE_GUIDANCE']
valid_report_formats = ["CSV", "JSON"]
def create_version_reports(self, version, report_list, format="CSV"):
    assert all(list(map(lambda k: k in valid_categories, report_list))), "One or more selected report categories in {} are not valid ({})".format(
        report_list, valid_categories)
    assert format in valid_report_formats, "Format must be one of {}".format(valid_report_formats)

    post_data = {
        'categories': report_list,
        'versionId': version['_meta']['href'].split("/")[-1],
        'reportType': 'VERSION',
        'reportFormat': format
    }
    version_reports_url = self.get_link(version, 'versionReport')
    return self.execute_post(version_reports_url, post_data)

valid_notices_formats = ["TEXT", "JSON"]
def create_version_notices_report(self, version, format="TEXT", include_copyright_info=True):
    assert format in valid_notices_formats, "Format must be one of {}".format(valid_notices_formats)

    post_data = {
        'versionId': object_id(version),
        'reportType': 'VERSION_LICENSE',
        'reportFormat': format
    }
    if include_copyright_info:
        post_data.update({'categories': ["COPYRIGHT_TEXT"] })

    notices_report_url = self.get_link(version, 'licenseReports')
    return self.execute_post(notices_report_url, post_data)

def download_report(self, report_id):
    # TODO: Fix me, looks like the reports should be downloaded from different paths than the one here, and depending on the type and format desired the path can change
    url = self.get_urlbase() + "/api/reports/{}".format(report_id)
    return self.execute_get(url, {'Content-Type': 'application/zip', 'Accept':'application/zip'})

def download_notification_report(self, report_location_url):
    '''Download the notices report using the report URL. Inspect the report object to determine
    the format and use the appropriate media header'''
    custom_headers = {'Accept': 'application/vnd.blackducksoftware.report-4+json'}
    response = self.execute_get(report_location_url, custom_headers=custom_headers)
    report_obj = response.json()

    if report_obj['reportFormat'] == 'TEXT':
        download_url = self.get_link(report_obj, "download") + ".json"
        logger.debug("downloading report from {}".format(download_url))
        response = self.execute_get(download_url, {'Accept': 'application/zip'})
    else:
        # JSON
        contents_url = self.get_link(report_obj, "content")
        logger.debug("retrieving report contents from {}".format(contents_url))
        response = self.execute_get(contents_url, {'Accept': 'application/json'})
    return response, report_obj['reportFormat']

##
#
# (Global) Vulnerability reports
#
##
valid_vuln_status_report_formats = ["CSV", "JSON"]
def create_vuln_status_report(self, format="CSV"):
    assert format in valid_vuln_status_report_formats, "Format must be one of {}".format(valid_vuln_status_report_formats)

    post_data = {
        "reportFormat": format,
        "locale": "en_US"
    }
    url = self.get_apibase() + "/vulnerability-status-reports"
    custom_headers = {
        'Content-Type': 'application/vnd.blackducksoftware.report-4+json',
        'Accept': 'application/vnd.blackducksoftware.report-4+json'
    }
    return self.execute_post(url, custom_headers=custom_headers, data=post_data)
